@@ -7,10 +7,18 @@ use std::fmt;
 
 use tls_api::Error;
 use tls_api::Result;
+use openssl::ssl::AlpnError;
 
 
-pub struct TlsConnectorBuilder(pub openssl::ssl::SslConnectorBuilder);
-pub struct TlsConnector(pub openssl::ssl::SslConnector);
+pub struct TlsConnectorBuilder {
+    pub builder: openssl::ssl::SslConnectorBuilder,
+    pub verify_hostname: bool,
+}
+
+pub struct TlsConnector {
+    pub connector: openssl::ssl::SslConnector,
+    pub verify_hostname: bool,
+}
 
 pub struct TlsAcceptorBuilder(pub openssl::ssl::SslAcceptorBuilder);
 pub struct TlsAcceptor(pub openssl::ssl::SslAcceptor);
@@ -23,13 +31,32 @@ pub const HAS_ALPN: bool = true;
 pub const HAS_ALPN: bool = false;
 
 
+fn encode_alpn_protos(protos: &[&[u8]]) -> Result<Vec<u8>> {
+    let mut r = Vec::new();
+    for proto in protos {
+        if proto.len() > 255 {
+            return Err(Error::new_other("prototol len"));
+        }
+        r.push(proto.len() as u8);
+        r.extend_from_slice(proto);
+    }
+    Ok(r)
+}
+
+#[cfg(test)]
+#[test]
+fn test_encode_alpn_protos() {
+    assert_eq!(&b"\x06spdy/1\x08http/1.1"[..], &encode_alpn_protos(&[b"spdy/1", b"http/1.1"]).unwrap()[..]);
+}
+
+
 impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
     type Connector = TlsConnector;
 
     type Underlying = openssl::ssl::SslConnectorBuilder;
 
     fn underlying_mut(&mut self) -> &mut openssl::ssl::SslConnectorBuilder {
-        &mut self.0
+        &mut self.builder
     }
 
     fn supports_alpn() -> bool {
@@ -38,7 +65,7 @@ impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
 
     #[cfg(has_alpn)]
     fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
-        self.0.set_alpn_protocols(protocols).map_err(Error::new)
+        self.builder.set_alpn_protos(&encode_alpn_protos(protocols)?).map_err(Error::new)
     }
 
     #[cfg(not(has_alpn))]
@@ -46,11 +73,20 @@ impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
         Err(Error::new_other("openssl is compiled without alpn"))
     }
 
-    fn add_root_certificate(&mut self, cert: tls_api::Certificate) -> Result<&mut Self> {
-        let cert = openssl::x509::X509::from_der(&cert.into_der())
-            .map_err(Error::new)?;
+    fn set_verify_hostname(&mut self, verify: bool) -> Result<()> {
+        self.verify_hostname = verify;
+        Ok(())
+    }
 
-        self.0
+    fn add_root_certificate(&mut self, cert: tls_api::Certificate) -> Result<&mut Self> {
+        let cert = match cert.format {
+            tls_api::CertificateFormat::DER => openssl::x509::X509::from_der(&cert.bytes)
+                .map_err(Error::new)?,
+            tls_api::CertificateFormat::PEM => openssl::x509::X509::from_pem(&cert.bytes)
+                .map_err(Error::new)?,
+        };
+
+        self.builder
             .cert_store_mut()
             .add_cert(cert)
             .map_err(Error::new)?;
@@ -59,13 +95,16 @@ impl tls_api::TlsConnectorBuilder for TlsConnectorBuilder {
     }
 
     fn build(self) -> Result<TlsConnector> {
-        Ok(TlsConnector(self.0.build()))
+        Ok(TlsConnector {
+            connector: self.builder.build(),
+            verify_hostname: self.verify_hostname,
+        })
     }
 }
 
 impl TlsConnectorBuilder {
     pub fn builder_mut(&mut self) -> &mut openssl::ssl::SslConnectorBuilder {
-        &mut self.0
+        &mut self.builder
     }
 }
 
@@ -95,12 +134,9 @@ impl<S : io::Read + io::Write + fmt::Debug> io::Write for TlsStream<S> {
 impl<S : io::Read + io::Write + fmt::Debug + Send + Sync + 'static> tls_api::TlsStreamImpl<S> for TlsStream<S> {
     fn shutdown(&mut self) -> io::Result<()> {
         match self.0.shutdown() {
-            Ok(_) |
-            Err(openssl::ssl::Error::ZeroReturn) => Ok(()),
-            Err(openssl::ssl::Error::Stream(e)) |
-            Err(openssl::ssl::Error::WantRead(e)) |
-            Err(openssl::ssl::Error::WantWrite(e)) => Err(e),
-            Err(e) => return Err(io::Error::new(io::ErrorKind::Other, e)),
+            Ok(_) => Ok(()),
+            Err(ref e) if e.code() == openssl::ssl::ErrorCode::ZERO_RETURN => Ok(()),
+            Err(e) => Err(io::Error::new(io::ErrorKind::Other, e)),
         }
     }
 
@@ -148,12 +184,12 @@ fn map_handshake_error<S>(e: openssl::ssl::HandshakeError<S>) -> tls_api::Handsh
 {
     match e {
         openssl::ssl::HandshakeError::SetupFailure(e) => {
-            tls_api::HandshakeError::Failure(Error::new(openssl::ssl::Error::Ssl(e)))
+            tls_api::HandshakeError::Failure(Error::new(e))
         }
         openssl::ssl::HandshakeError::Failure(e) => {
             tls_api::HandshakeError::Failure(Error::new(e.into_error()))
         },
-        openssl::ssl::HandshakeError::Interrupted(s) => {
+        openssl::ssl::HandshakeError::WouldBlock(s) => {
             tls_api::HandshakeError::Interrupted(
                 tls_api::MidHandshakeTlsStream::new(MidHandshakeTlsStream(Some(s))))
         }
@@ -164,27 +200,23 @@ impl tls_api::TlsConnector for TlsConnector {
     type Builder = TlsConnectorBuilder;
 
     fn builder() -> Result<TlsConnectorBuilder> {
-        openssl::ssl::SslConnectorBuilder::new(openssl::ssl::SslMethod::tls())
-            .map(TlsConnectorBuilder)
-            .map_err(Error::new)
+        let builder = openssl::ssl::SslConnector::builder(openssl::ssl::SslMethod::tls())
+            .map_err(Error::new)?;
+        Ok(TlsConnectorBuilder {
+            builder,
+            verify_hostname: true,
+        })
     }
 
     fn connect<S>(&self, domain: &str, stream: S)
         -> result::Result<tls_api::TlsStream<S>, tls_api::HandshakeError<S>>
             where S : io::Read + io::Write + fmt::Debug + Send + Sync + 'static
     {
-        self.0.connect(domain, stream)
-            .map(|s| tls_api::TlsStream::new(TlsStream(s)))
-            .map_err(map_handshake_error)
-    }
-
-    fn danger_connect_without_providing_domain_for_certificate_verification_and_server_name_indication<S>(
-        &self,
-        stream: S)
-        -> result::Result<tls_api::TlsStream<S>, tls_api::HandshakeError<S>>
-            where S : io::Read + io::Write + fmt::Debug + Send + Sync + 'static
-    {
-        self.0.danger_connect_without_providing_domain_for_certificate_verification_and_server_name_indication(stream)
+        self.connector
+            .configure().map_err(|e| tls_api::HandshakeError::Failure(Error::new(e)))?
+            .verify_hostname(self.verify_hostname)
+            .use_server_name_indication(false)
+            .connect(domain, stream)
             .map(|s| tls_api::TlsStream::new(TlsStream(s)))
             .map_err(map_handshake_error)
     }
@@ -198,13 +230,18 @@ impl TlsAcceptorBuilder {
         let pkcs12 = openssl::pkcs12::Pkcs12::from_der(pkcs12).map_err(Error::new)?;
         let pkcs12 = pkcs12.parse(password).map_err(Error::new)?;
 
-        openssl::ssl::SslAcceptorBuilder::mozilla_intermediate(
-            openssl::ssl::SslMethod::tls(),
-            &pkcs12.pkey,
-            &pkcs12.cert,
-            &pkcs12.chain)
-                .map(TlsAcceptorBuilder)
-                .map_err(Error::new)
+        let mut builder = openssl::ssl::SslAcceptor::mozilla_intermediate(
+            openssl::ssl::SslMethod::tls()).map_err(Error::new)?;
+
+        if let Some(chain) = pkcs12.chain {
+            for x509 in chain {
+                builder.add_extra_chain_cert(x509).map_err(Error::new)?;
+            }
+        }
+        builder.set_certificate(&pkcs12.cert).map_err(Error::new)?;
+        builder.set_private_key(&pkcs12.pkey).map_err(Error::new)?;
+
+        Ok(TlsAcceptorBuilder(builder))
     }
 }
 
@@ -223,7 +260,14 @@ impl tls_api::TlsAcceptorBuilder for TlsAcceptorBuilder {
 
     #[cfg(has_alpn)]
     fn set_alpn_protocols(&mut self, protocols: &[&[u8]]) -> Result<()> {
-        self.0.set_alpn_protocols(protocols).map_err(Error::new)
+        let protocols = encode_alpn_protos(protocols)?;
+        self.0.set_alpn_select_callback(move |_ssl, client_protocols| {
+            match openssl::ssl::select_next_proto(&protocols, client_protocols) {
+                Some(selected) => Ok(selected),
+                None => Err(AlpnError::NOACK),
+            }
+        });
+        Ok(())
     }
 
     #[cfg(not(has_alpn))]
